@@ -41,7 +41,7 @@
 #include <QThread>
 
 #if QT_CONFIG(opengl)
-#include <QtGui/private/qshaderdescription_p_p.h>
+#include <QtGui/private/qshaderdescription_p.h>
 #endif
 
 #ifndef GL_TEXTURE_EXTERNAL_OES
@@ -453,9 +453,8 @@ private:
  * Constructs a QWaylandQuickItem with the given \a parent.
  */
 QWaylandQuickItem::QWaylandQuickItem(QQuickItem *parent)
-    : QQuickItem(*new QWaylandQuickItemPrivate(), parent)
+    : QWaylandQuickItem(*new QWaylandQuickItemPrivate(), parent)
 {
-    d_func()->init();
 }
 
 /*!
@@ -465,6 +464,7 @@ QWaylandQuickItem::QWaylandQuickItem(QWaylandQuickItemPrivate &dd, QQuickItem *p
     : QQuickItem(dd, parent)
 {
     d_func()->init();
+    connect(this, &QQuickItem::activeFocusChanged, this, &QWaylandQuickItem::updateFocus);
 }
 
 /*!
@@ -474,13 +474,16 @@ QWaylandQuickItem::~QWaylandQuickItem()
 {
     Q_D(QWaylandQuickItem);
     disconnect(this, &QQuickItem::windowChanged, this, &QWaylandQuickItem::updateWindow);
+    disconnect(this, &QQuickItem::activeFocusChanged, this, &QWaylandQuickItem::updateFocus);
     QMutexLocker locker(d->mutex);
-    if (d->provider)
+    if (d->provider) {
+        disconnect(d->texProviderConnection);
         d->provider->deleteLater();
+    }
 }
 
 /*!
- * \qmlproperty WaylandCompositor QtWaylandCompositor::WaylandQuickItem::compositor
+ * \qmlproperty WaylandCompositor QtWayland.Compositor::WaylandQuickItem::compositor
  *
  * This property holds the compositor for the surface rendered by this WaylandQuickItem.
  */
@@ -506,7 +509,7 @@ QWaylandView *QWaylandQuickItem::view() const
 }
 
 /*!
- * \qmlproperty WaylandSurface QtWaylandCompositor::WaylandQuickItem::surface
+ * \qmlproperty WaylandSurface QtWayland.Compositor::WaylandQuickItem::surface
  *
  * This property holds the surface rendered by this WaylandQuickItem.
  */
@@ -534,11 +537,13 @@ void QWaylandQuickItem::setSurface(QWaylandSurface *surface)
         emit compositorChanged();
     if (oldSurf != surface)
         emit surfaceChanged();
+
+    updateFocus();
     update();
 }
 
 /*!
- * \qmlproperty enum QtWaylandCompositor::WaylandQuickItem::origin
+ * \qmlproperty enum QtWayland.Compositor::WaylandQuickItem::origin
  *
  * This property holds the origin of the QWaylandQuickItem.
  */
@@ -757,7 +762,7 @@ void QWaylandQuickItem::keyPressEvent(QKeyEvent *event)
 void QWaylandQuickItem::keyReleaseEvent(QKeyEvent *event)
 {
     Q_D(QWaylandQuickItem);
-    if (d->shouldSendInputEvents() && hasFocus()) {
+    if (d->shouldSendInputEvents()) {
         QWaylandSeat *seat = compositor()->seatFor(event);
         seat->sendFullKeyEvent(event);
     } else {
@@ -894,8 +899,15 @@ void QWaylandQuickItem::handlePlaceBelow(QWaylandSurface *referenceSurface)
     }
 }
 
+void QWaylandQuickItem::updateFocus()
+{
+    Q_D(const QWaylandQuickItem);
+    if (hasActiveFocus() && compositor())
+        compositor()->defaultSeat()->setKeyboardFocus(d->view->surface());
+}
+
 /*!
-  \qmlproperty object QtWaylandCompositor::WaylandQuickItem::subsurfaceHandler
+  \qmlproperty object QtWayland.Compositor::WaylandQuickItem::subsurfaceHandler
 
   This property provides a way to override the default subsurface behavior.
 
@@ -933,6 +945,11 @@ void QWaylandQuickItem::setSubsurfaceHandler(QObject *handler)
     }
 }
 
+/*!
+ * \qmlproperty WaylandOutput QtWayland.Compositor::WaylandQuickItem::output
+ *
+ * This property holds the output on which this item is displayed.
+ */
 /*!
  * \property QWaylandQuickItem::output
  *
@@ -951,7 +968,7 @@ void QWaylandQuickItem::setOutput(QWaylandOutput *output)
 }
 
 /*!
- * \qmlproperty bool QtWaylandCompositor::WaylandQuickItem::bufferLocked
+ * \qmlproperty bool QtWayland.Compositor::WaylandQuickItem::bufferLocked
  *
  * This property holds whether the item's buffer is currently locked. As long as
  * the buffer is locked, it will not be released and returned to the client.
@@ -1166,7 +1183,7 @@ void QWaylandQuickItem::updateSize()
 }
 
 /*!
- * \qmlproperty bool QtWaylandCompositor::WaylandQuickItem::focusOnClick
+ * \qmlproperty bool QtWayland.Compositor::WaylandQuickItem::focusOnClick
  *
  * This property specifies whether the WaylandQuickItem should take focus when
  * it is clicked or touched.
@@ -1281,7 +1298,7 @@ QVariant QWaylandQuickItem::inputMethodQuery(Qt::InputMethodQuery query, QVarian
 #endif
 
 /*!
-    \qmlproperty bool QtWaylandCompositor::WaylandQuickItem::paintEnabled
+    \qmlproperty bool QtWayland.Compositor::WaylandQuickItem::paintEnabled
 
     Returns true if the item is hidden, though the texture
     is still updated. As opposed to hiding the item by
@@ -1316,7 +1333,7 @@ void QWaylandQuickItem::setPaintEnabled(bool enabled)
 }
 
 /*!
-    \qmlproperty  bool QtWaylandCompositor::WaylandQuickItem::touchEventsEnabled
+    \qmlproperty  bool QtWayland.Compositor::WaylandQuickItem::touchEventsEnabled
 
     This property holds \c true if touch events are forwarded to the client
     surface, \c false otherwise.
@@ -1411,7 +1428,7 @@ void QWaylandQuickItem::updateInputMethod(Qt::InputMethodQueries queries)
 #endif
 
 /*!
- * \qmlsignal void QtWaylandCompositor::WaylandQuickItem::surfaceDestroyed()
+ * \qmlsignal void QtWayland.Compositor::WaylandQuickItem::surfaceDestroyed()
  *
  * This signal is emitted when the client has destroyed the \c wl_surface associated
  * with the WaylandQuickItem. The handler for this signal is expected to either destroy the
@@ -1478,8 +1495,24 @@ QSGNode *QWaylandQuickItem::updatePaintNode(QSGNode *oldNode, UpdatePaintNodeDat
             d->newTexture = true;
         }
 
-        if (!d->provider)
+        if (!d->provider) {
             d->provider = new QWaylandSurfaceTextureProvider();
+            if (compositor()) {
+                d->texProviderConnection =
+                    QObject::connect(
+                            compositor(),
+                            &QObject::destroyed,
+                            this,
+                            [this](QObject*) {
+                                    auto *itemPriv = QWaylandQuickItemPrivate::get(this);
+                                    if (itemPriv->provider) {
+                                        itemPriv->provider->deleteLater();
+                                        itemPriv->provider = nullptr;
+                                    }
+                                    disconnect(itemPriv->texProviderConnection); }
+                    );
+            }
+        }
 
         if (d->newTexture) {
             d->newTexture = false;
